from __future__ import annotations

import itertools

class Piece:
    ''' Basic functionality of TicTacToe pieces. '''

    piece_symbols = '-XOVTSZLC'  # 0 is hardcoded to be empty piece
    
    def __init__(self, code: int = 0) -> None:
        ''' Initialize piece. '''
        self.code = code

    @staticmethod
    def next(piece: Piece, players: int) -> Piece:
        ''' Return the next piece that will be placed on board. '''
        if piece.code == players:
            return Piece(1)
        else:
            return Piece(piece.code+1)

    @staticmethod
    def previous(piece: Piece, players: int) -> Piece:
        ''' Return the previous piece that was placed on board. '''
        if piece.code == 1:
            return Piece(players)
        else:
            return Piece(piece.code-1)

    def __eq__(self, other: Piece) -> bool:
        return self.code == other.code

    def __int__(self) -> int:
        return self.code

    def __repr__(self) -> str:
        return Piece.piece_symbols[self.code]

class Board:
    '''
    Only basic functionality of Board. There is no use of Tic Tac Toe game
    logic like win/draw, game ending, turn etc. Game logic is implemented in
    child class TicTacToe.
    '''

    def __init__(self, m: int = 3, n: int = 3, infinite: bool = False) -> None:
        '''
        Initialize a board.
        (m, n) are dimensions of the board: n is number of rows, m is number of columns.
        infinite makes board an infinite matrix for with non-negative indexes (not actually infinite, but with dynamically changing dimensions).
        '''
        self.infinite = infinite
        self.m = m
        self.n = n
        self.board = [[Piece()]*m for _ in range(n)]

    def increase_dimensions(self, to_dim: int) -> None:
        ''' Scale up dimensions to (s, s). '''
        newboard = Board(to_dim, to_dim, infinite=True)

        for (x, y) in newboard.moves_coords():
            if (x, y) in self.moves_coords():
                newboard[x, y] = self[x, y]
            else:
                newboard[x, y] = Piece()

        self.n, self.m = to_dim, to_dim
        self.board = newboard.board.copy()

    def columns(self) -> list:
        ''' Return columns for checking win. '''
        board = self.board
        return [list(field) for field in zip(*board)]

    def diagonals(self) -> list:
        ''' Return diagonals for checking win. '''
        diags = []
        for sm in range(self.m+self.n-1):
            strip = []
            for (i, j) in self.moves_coords():
                if i+j == sm:
                    strip.append(self[i, j])
            diags.append(strip)
        return diags

    def antidiagonals(self) -> list:
        ''' Return diagonals for checking win. '''
        # Rotating board by 90 degree and then getting its diagonals
        rotated_board = Board(m=self.n, n=self.m)
        rotated_board.board = list(zip(*self.board))[::-1]
        return rotated_board.diagonals()

    def lines(self) -> list:
        ''' Return all possible lines to be winned. '''
        return self.board + self.columns() + self.diagonals() + self.antidiagonals()

    def flat(self) -> list:
        ''' Return flat format of board. '''
        board = self.board
        return sum(board, [])
    
    def moves_coords(self) -> list:
        ''' Return all the fields on board'''
        return [(i,j) for (i,j) in itertools.product(range(self.n), range(self.m))]

    def __getitem__(self, key: tuple) -> None:
        if not key in self.moves_coords() and not self.infinite:
            raise IndexError('Board index out of range.')

        if not key in self.moves_coords() and self.infinite:
            to_dim = max(key[0]+1, key[1]+1)
            self.increase_dimensions(to_dim)

        return self.board[key[0]][key[1]]

    def __setitem__(self, key: tuple, value: Piece) -> None:
        if not key in self.moves_coords() and not self.infinite:
            raise IndexError('Board index out of range.')

        if not key in self.moves_coords() and self.infinite:
            to_dim = max(key[0]+1, key[1]+1)
            self.increase_dimensions(to_dim)
  
        self.board[key[0]][key[1]] = value

    def __repr__(self) -> str:
        printed = []
        for row in self.board:
            for field in row:
                printed.append(str(field))
            printed.append('\n')
        return ''.join(printed)