from __future__ import annotations

from pytactoe.board import Piece, Board


class TicTacToe(Board):
    ''' Game logic of TicTacToe. '''
    def __init__(self, m: int = 3, n: int = 3, k: int = 3, infinite: bool = False, first_player: Piece = Piece(1), players: int = 2) -> None:
        ''' Initialize an (m, n, k)-game. '''
        super().__init__(m, n, infinite)
        self.k = k
        self.turn = first_player
        self.players = players
        if not 2 <= players <= 8:
            raise ValueError('Number of players must be an int from 2 to 8.')
        if not 1 <= first_player.code <= players:
            raise ValueError('Invalid first player, first_player code must be from 1 to players.')
        self.ended = False
        self.draw = False
        self.winned = False
        self.winner = Piece()
        self.move_count = 0
        self.move_history = []

    def copy(self, game: TicTacToe) -> None:
        self.__dict__ = game.__dict__.copy()

    def checkdraw(self) -> bool:
        ''' Return True and update attributes if draw. '''
        if self.winned == True:
            return False
        if Piece() not in self.flat():
            self.ended = True
            self.draw = True
            return True
        return False

    @staticmethod
    def check_line_winned(line: list, k=3) -> bool:
        ''' Return True if line contain k-in-a-row. '''
        if k > len(line):
            return False

        for i in range(len(line)-k+1):
            next_k_fields = line[i:i+k]
            if Piece() in next_k_fields:
                continue
            if all(field==next_k_fields[0] for field in next_k_fields):
                return True

        return False

    def checkwin(self) -> bool:
        ''' Return True and update attributes if win. '''
        winned = False
        for line in self.lines():
            winned = TicTacToe.check_line_winned(line, self.k)
            if winned:
                self.winner = self.turn
                self.winned = True
                self.ended = True
                break
        return winned

    def coords_winnedline(self) -> list:
        ''' Return coordinates of a winned line. The line is chosen as the first in self.lines() list.  '''
        # Unimplemented
        pass
        
    def changeturn(self) -> None:
        ''' Change turn attribute preparing for the next move. '''
        self.turn = Piece.next(self.turn, self.players)

    def revertturn(self) -> None:
        ''' Revert turn attribute. '''
        self.turn = Piece.previous(self.turn, self.players)
    
    def movevalid(self, x: int, y: int) -> bool:
        ''' Return True if move is valid. '''
        if not self.ended:
            return self[x, y] == Piece()
        return False

    def move(self, x: int, y: int) -> bool:
        ''' Make move and update attributes. Return True if move is successful. '''
        if self.movevalid(x, y):
            self[x, y] = self.turn
            self.checkwin()
            self.checkdraw()
            self.changeturn()
            self.move_history.append((x,y))
            self.move_count += 1
            return True
        return False

    def pop(self) -> None:
        ''' Revert board to the previous position. '''
        if not self.move_count:
            return

        last_move = self.move_history[-1]

        self.draw = False
        self.ended = False
        self.winned = False
        self.winner = Piece()
        self.move_count -= 1
        self.move_history.pop()
        self[last_move[0], last_move[1]] = Piece()

        self.revertturn()
        self.checkwin()
        self.checkdraw()

    def legalmoves(self) -> list:
        ''' Return list of tuples of all legal move coordinates. '''
        return [(i,j) for (i,j) in self.moves_coords() if self.movevalid(i,j)]

    def __repr__(self) -> str:
        return super().__repr__() + f'(m,n,k)=({self.m},{self.n},{self.k}), infinite={self.infinite}, '\
            f'turn={self.turn}, ended={self.ended}, winned={self.winned}, winner={self.winner}, draw={self.draw}'